#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import functools
import sys
from multiprocessing.dummy import Pool as ThreadPool

import redis
from redis.client import Lock
from redis.sentinel import Sentinel

from .commands import SHARD_METHODS
from ._compat import basestring, iteritems
from .hashring import HashRing
from .helpers import format_servers
from .pipeline import Pipeline
from .sentinel import SentinelRedis

_findhash = re.compile('.*\{(.*)\}.*', re.I)

_PYTHON3 = True if sys.version_info > (3, 0, 0) else False

def list_or_args(keys, args):
    # returns a single list combining keys and args
    try:
        iter(keys)
        # a string can be iterated, but indicates
        # keys wasn't passed as a list
        if isinstance(keys, basestring):
            keys = [keys]
    except TypeError:
        keys = [keys]
    if args:
        keys.extend(args)
    return keys


class RedisShardAPI(object):

    def __init__(self, servers, hash_method='crc32', sentinel=None, strict_redis=False):
        self.nodes = []
        self.connections = {}
        self.pool = None
        servers = format_servers(servers)
        if sentinel:
            sentinel = Sentinel(sentinel['hosts'], socket_timeout=sentinel.get('socket_timeout', 1))
        for server_config in servers:
            name = server_config.pop('name')
            if name in self.connections:
                raise ValueError("server's name config must be unique")
            if sentinel:
                self.connections[name] = SentinelRedis(sentinel, name)
            elif strict_redis:
                self.connections[name] = redis.StrictRedis(**server_config)
            else:
                self.connections[name] = redis.Redis(**server_config)
            server_config['name'] = name
            self.nodes.append(name)
        self.ring = HashRing(self.nodes, hash_method=hash_method)

    def get_server_name(self, key):
        if not _PYTHON3 and not isinstance(key, bytes):
            g = _findhash.match(key)
            if g is not None and len(g.groups()) > 0:
                key = g.groups()[0]
        name = self.ring.get_node(key)
        return name

    def get_server(self, key):
        name = self.get_server_name(key)
        return self.connections[name]

    def _build_pool(self):
        if self.pool is None:
            self.pool = ThreadPool(len(self.nodes))

    def __wrap(self, method, *args, **kwargs):
        try:
            key = args[0]
            assert isinstance(key, basestring)
        except:
            raise ValueError("method '%s' requires a key param as the first argument" % method)
        server = self.get_server(key)
        f = getattr(server, method)
        return f(*args, **kwargs)

    def __wrap_eval(self, method, script_or_sha, numkeys, *keys_and_args):
        if numkeys != 1:
            raise NotImplementedError("The key must be single string;mutiple keys cannot be sharded")
        key = keys_and_args[0]
        server = self.get_server(key)
        f = getattr(server, method)
        return f(script_or_sha, numkeys, *keys_and_args)

    def __wrap_tag(self, method, *args, **kwargs):
        key = args[0]
        if isinstance(key, basestring) and '{' in key:
            server = self.get_server(key)
        elif isinstance(key, list) and '{' in key[0]:
            server = self.get_server(key[0])
        else:
            raise ValueError("method '%s' requires tag key params as its arguments" % method)
        method = method.lstrip("tag_")
        f = getattr(server, method)
        return f(*args, **kwargs)

    def __getattr__(self, method):
        if method in SHARD_METHODS:
            return functools.partial(self.__wrap, method)
        elif method in ('eval', 'evalsha'):
            return functools.partial(self.__wrap_eval, method)
        elif method.startswith("tag_"):
            return functools.partial(self.__wrap_tag, method)
        else:
            raise NotImplementedError("method '%s' cannot be sharded" % method)

    #########################################
    ###  some methods implement as needed ###
    ########################################
    def brpop(self, key, timeout=0):
        if not isinstance(key, basestring):
            raise NotImplementedError("The key must be single string;mutiple keys cannot be sharded")
        server = self.get_server(key)
        return server.brpop(key, timeout)

    def blpop(self, key, timeout=0):
        if not isinstance(key, basestring):
            raise NotImplementedError("The key must be single string;mutiple keys cannot be sharded")
        server = self.get_server(key)
        return server.blpop(key, timeout)

    def keys(self, key):
        _keys = []
        for node in self.nodes:
            server = self.connections[node]
            _keys.extend(server.keys(key))
        return _keys

    def mget(self, keys, *args):
        """
        Returns a list of values ordered identically to ``keys``
        """
        args = list_or_args(keys, args)
        server_keys = {}
        ret_dict = {}
        for key in args:
            server_name = self.get_server_name(key)
            server_keys[server_name] = server_keys.get(server_name, [])
            server_keys[server_name].append(key)
        for server_name, sub_keys in iteritems(server_keys):
            values = self.connections[server_name].mget(sub_keys)
            ret_dict.update(dict(zip(sub_keys, values)))
        result = []
        for key in args:
            result.append(ret_dict.get(key, None))
        return result

    def mset(self, mapping):
        """
        Sets each key in the ``mapping`` dict to its corresponding value
        """
        servers = {}
        for key, value in mapping.items():
            server_name = self.get_server_name(key)
            servers.setdefault(server_name, [])
            servers[server_name].append((key, value))
        for name, items in servers.items():
            self.connections[name].mset(dict(items))
        return True

    def flushdb(self):
        for node in self.nodes:
            server = self.connections[node]
            server.flushdb()

    def lock(self, name, timeout=None, sleep=0.1):
        """
        Return a new Lock object using key ``name`` that mimics
        the behavior of threading.Lock.

        If specified, ``timeout`` indicates a maximum life for the lock.
        By default, it will remain locked until release() is called.

        ``sleep`` indicates the amount of time to sleep per loop iteration
        when the lock is in blocking mode and another client is currently
        holding the lock.
        """
        return Lock(self, name, timeout=timeout, sleep=sleep)

    def pipeline(self):
        return Pipeline(self)

    def script_load(self, script):
        shas = []
        for node in self.nodes:
            server = self.connections[node]
            shas.append(server.script_load(script))
        if not all(x == shas[0] for x in shas):
            raise ValueError('not all server returned same sha')
        return shas[0]

    def haskey(self, key):
        server_name = self.get_server_name(key)
        return key in self.connections[server_name]

    def __delitem__(self, key):
        server_name = self.get_server_name(key)
        del self.connections[server_name][key]



#!/usr/bin/env python
SHARD_METHODS = frozenset(['restore',
                           'debug object',
                           'renamenx',
                           'hincrby',
                           'zrevrank',
                           'zcount',
                           'move',
                           'pop',
                           'hexists',
                           'get_type',
                           'incrbyfloat',
                           'ttl',
                           'sinter',
                           'getbit',
                           'decr',
                           'rpush',
                           'pttl',
                           'zscore',
                           'delete',
                           'dump',
                           'zincrby',
                           'zrevrangebyscore',
                           'persist',
                           'zrem',
                           'zadd',
                           'linsert',
                           'sort',
                           'zcard',
                           'bitcount',
                           'get',
                           'setnx',
                           'watch',
                           'incrby',
                           'lindex',
                           'hget',
                           'hlen',
                           'expireat',
                           'type',
                           'zincr',
                           'lset',
                           'hkeys',
                           'setrange',
                           'del',
                           'hset',
                           'decrby',
                           'zrange',
                           'rename',
                           'set',
                           'exists',
                           'hincrbyfloat',
                           'pexpireat',
                           'hvals',
                           'setex',
                           'sdiff',
                           'blpop',
                           'strlen',
                           'append',
                           'srem',
                           'pexpire',
                           'getrange',
                           'zrevrange',
                           'zremrangebyscore',
                           'hsetnx',
                           'lpushx',
                           'rpushx',
                           'hmset',
                           'srandmember',
                           'sismember',
                           'getset',
                           'smembers',
                           'zrank',
                           'psetex',
                           'zremrangebyrank',
                           'sadd',
                           'ltrim',
                           'spop',
                           'incr',
                           'expire',
                           'brpop',
                           'lrange',
                           'sunion',
                           'setbit',
                           'zrangebyscore',
                           'rpop',
                           'lrem',
                           'lpop',
                           'hgetall',
                           'lpush',
                           'hmget',
                           'push',
                           'hdel',
                           'scard',
                           'llen',
                           'publish'])


READ_COMMANDS = frozenset([
    'info', 'smembers', 'hlen', 'hmget', 'srandmember', 'hvals', 'randomkey', 'strlen',
    'dbsize', 'keys', 'ttl', 'lindex', 'type', 'llen', 'dump', 'scard', 'echo', 'lrange',
    'zcount', 'exists', 'sdiff', 'zrange', 'mget', 'zrank', 'get', 'getbit', 'getrange',
    'zrevrange', 'zrevrangebyscore', 'hexists', 'object', 'sinter', 'zrevrank', 'hget',
    'zscore', 'hgetall', 'sismember'])



#!/usr/bin/env python
# encoding: utf-8
from redis import Redis
from .commands import READ_COMMANDS


class SentinelRedis(object):

    def __init__(self, sentinel, service_name):
        self.master = sentinel.master_for(service_name, redis_class=Redis)
        self.slave = sentinel.slave_for(service_name, redis_class=Redis)

    def __getattr__(self, method):
        if method in READ_COMMANDS:
            return getattr(self.slave, method)
        else:
            return getattr(self.master, method)




import functools
import logging

from .commands import SHARD_METHODS
from ._compat import basestring, dictvalues


class Pipeline(object):

    def __init__(self, shard_api):
        self.shard_api = shard_api
        self.pipelines = {}
        self.__counter = 0
        self.__indexes = {}
        self.__watching = None
        self.shard_api._build_pool()

    def get_pipeline(self, key):
        name = self.shard_api.get_server_name(key)
        if name not in self.pipelines:
            self.pipelines[name] = self.shard_api.connections[name].pipeline()
        return self.pipelines[name]

    def __record_index(self, pipeline):
        if pipeline not in self.__indexes:
            self.__indexes[pipeline] = [self.__counter]
        else:
            self.__indexes[pipeline].append(self.__counter)
        self.__counter += 1

    def __wrap(self, method, *args, **kwargs):
        try:
            key = args[0]
            assert isinstance(key, basestring)
        except:
            raise ValueError("method '%s' requires a key param as the first argument" % method)
        pipeline = self.get_pipeline(key)
        f = getattr(pipeline, method)
        r = f(*args, **kwargs)
        self.__record_index(pipeline)
        return r

    def __wrap_eval(self, method, script_or_sha, numkeys, *keys_and_args):
        if numkeys != 1:
            raise NotImplementedError("The key must be single string;mutiple keys cannot be sharded")
        key = keys_and_args[0]
        pipeline = self.get_pipeline(key)
        f = getattr(pipeline, method)
        r = f(script_or_sha, numkeys, *keys_and_args)
        self.__record_index(pipeline)
        return r

    def __wrap_tag(self, method, *args, **kwargs):
        key = args[0]
        if isinstance(key, basestring) and '{' in key:
            pipeline = self.get_pipeline(key)
        elif isinstance(key, list) and '{' in key[0]:
            pipeline = self.get_pipeline(key[0])
        else:
            raise ValueError("method '%s' requires tag key params as its arguments" % method)
        method = method.lstrip("tag_")
        f = getattr(pipeline, method)
        r = f(*args, **kwargs)
        self.__record_index(pipeline)
        return r

    def execute(self):
        try:
            if self.__watching:
                return self.pipelines[self.__watching].execute()
            else:
                results = []

                # Pipeline concurrently
                values = self.shard_api.pool.map(self.__unit_execute, dictvalues(self.pipelines))
                for v in values:
                    results.extend(v)

                self.__counter = 0
                self.__indexes = {}

                results.sort(key=lambda x: x[0])
                results = [r[1] for r in results]
                return results
        finally:
            self.reset()

    def __unit_execute(self, pipeline):
        result = pipeline.execute()
        return zip(self.__indexes.get(pipeline, []), result)

    def __getattr__(self, method):
        if method in SHARD_METHODS:
            return functools.partial(self.__wrap, method)
        elif method in ('eval', 'evalsha'):
            return functools.partial(self.__wrap_eval, method)
        elif method.startswith("tag_"):
            return functools.partial(self.__wrap_tag, method)
        else:
            raise NotImplementedError(
                "method '%s' cannot be pipelined" % method)

    def watch(self, *keys):
        if not keys:
            return

        watching_server_name = self.__watching
        if watching_server_name:
            for key in keys:
                name = self.shard_api.get_server_name(key)
                if watching_server_name != name:
                    raise ValueError('Cannot watch keys in different redis instances')
        else:
            for key in keys:
                name = self.shard_api.get_server_name(key)
                if watching_server_name is None:
                    watching_server_name = name
                elif watching_server_name != name:
                    raise ValueError('Cannot watch keys in different redis instances')

        pipeline = self.pipelines.get(watching_server_name)
        if not pipeline:
            pipeline = self.shard_api.connections[watching_server_name].pipeline()
            self.pipelines[watching_server_name] = pipeline
        r = pipeline.watch(*keys)
        self.__watching = watching_server_name
        return r

    def multi(self):
        if self.__watching:
            pipeline = self.pipelines[self.__watching]
            pipeline.multi()
        else:
            raise NotImplementedError('Cannot call MULTI in different redis instances')

    def reset(self):
        if self.__watching:
            pipeline = self.pipelines[self.__watching]
            try:
                pipeline.reset()
            except Exception:
                logging.warning('failed to reset pipeline')
            self.__watching = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def __del__(self):
        self.reset()



try:
    # python2
    from urllib import unquote
    from urlparse import urlparse
    from urlparse import parse_qsl
except ImportError:
    # python3
    from urllib.parse import unquote
    from urllib.parse import urlparse
    from urllib.parse import parse_qsl

from redis.connection import URL_QUERY_ARGUMENT_PARSERS


def _parse_url(url):
    scheme = urlparse(url).scheme
    schemeless = url[len(scheme) + 3:]
    # parse with HTTP URL semantics
    parts = urlparse('http://' + schemeless)

    # The first pymongo.Connection() argument (host) can be
    # a mongodb connection URI. If this is the case, don't
    # use port but let pymongo get the port(s) from the URI instead.
    # This enables the use of replica sets and sharding.
    # See pymongo.Connection() for more info.
    port = scheme != 'mongodb' and parts.port or None
    hostname = schemeless if scheme == 'mongodb' else parts.hostname
    path = parts.path or ''
    path = path[1:] if path and path[0] == '/' else path
    return (scheme, unquote(hostname or '') or None, port,
            unquote(parts.username or '') or None,
            unquote(parts.password or '') or None,
            unquote(path or '') or None,
            dict(parse_qsl(parts.query)))


def parse_url(url):
    scheme, host, port, user, password, db, query = _parse_url(url)
    assert scheme == 'redis'
    for param, value in query.items():
        if param in URL_QUERY_ARGUMENT_PARSERS:
            parser = URL_QUERY_ARGUMENT_PARSERS[param]
            query[param] = parser(value)
    return dict(host=host, port=port, password=password, db=db, **query)

if __name__ == '__main__':
    print(parse_url('redis://127.0.0.1:6379/1?name=node2&max_connections=100'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 consistent hashing for nosql client
 based on ezmobius client (redis-rb)
 and see this article http://amix.dk/blog/viewEntry/19367
"""

import zlib
import bisect
from hashlib import md5, sha1

from ._compat import xrange, b, long


hash_methods = {
    'crc32': lambda x: zlib.crc32(x) & 0xffffffff,
    'md5': lambda x: long(md5(x).hexdigest(), 16),
    'sha1': lambda x: long(sha1(x).hexdigest(), 16),
}


class HashRing(object):

    """Consistent hash for nosql API"""

    def __init__(self, nodes=[], replicas=128, hash_method='crc32'):
        """Manages a hash ring.

        `nodes` is a list of objects that have a proper __str__ representation.
        `replicas` indicates how many virtual points should be used pr. node,
        replicas are required to improve the distribution.
        `hash_method` is the key generator method.
        """
        self.hash_method = hash_methods[hash_method]
        self.nodes = []
        self.replicas = replicas
        self.ring = {}
        self.sorted_keys = []

        for n in nodes:
            self.add_node(n)

    def add_node(self, node):
        """Adds a `node` to the hash ring (including a number of replicas).
        """
        self.nodes.append(node)
        for x in xrange(self.replicas):
            ring_key = self.hash_method(b("%s:%d" % (node, x)))
            self.ring[ring_key] = node
            self.sorted_keys.append(ring_key)

        self.sorted_keys.sort()

    def remove_node(self, node):
        """Removes `node` from the hash ring and its replicas.
        """
        self.nodes.remove(node)
        for x in xrange(self.replicas):
            ring_key = self.hash_method(b("%s:%d" % (node, x)))
            self.ring.pop(ring_key)
            self.sorted_keys.remove(ring_key)

    def get_node(self, key):
        """Given a string key a corresponding node in the hash ring is returned.

        If the hash ring is empty, `None` is returned.
        """
        n, i = self.get_node_pos(key)
        return n

    def get_node_pos(self, key):
        """Given a string key a corresponding node in the hash ring is returned
        along with it's position in the ring.

        If the hash ring is empty, (`None`, `None`) is returned.
        """
        if len(self.ring) == 0:
            return [None, None]
        crc = self.hash_method(b(key))
        idx = bisect.bisect(self.sorted_keys, crc)
        # prevents out of range index
        idx = min(idx, (self.replicas * len(self.nodes)) - 1)
        return [self.ring[self.sorted_keys[idx]], idx]

    def iter_nodes(self, key):
        """Given a string key it returns the nodes as a generator that can hold the key.
        """
        if len(self.ring) == 0:
            yield None, None
        node, pos = self.get_node_pos(key)
        for k in self.sorted_keys[pos:]:
            yield k, self.ring[k]

    def __call__(self, key):
        return self.get_node(key)



# -*- coding: utf-8 -*-

import os
from setuptools import setup

import redis_shard


def read_file(*path):
    base_dir = os.path.dirname(__file__)
    file_path = (base_dir, ) + tuple(path)
    return open(os.path.join(*file_path)).read()


setup(
    name="redis-shard",
    url="https://github.com/zhihu/redis-shard",
    license="BSD License",
    author="Zhihu Inc.",
    author_email="opensource@zhihu.com",
    maintainer="Young King",
    maintainer_email="y@zhihu.com",
    description="Redis Sharding API",
    long_description=(
        read_file("README.rst") + "\n\n" +
        "Change History\n" +
        "===============\n\n" +
        read_file("CHANGES.rst")),
    version=redis_shard.__version__,
    packages=["redis_shard"],
    include_package_data=True,
    zip_safe=False,
    install_requires=['redis>=3.0.0'],
    tests_require=['Nose'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import sys
from .url import parse_url


def format_servers(servers):
    """
    :param servers: server list, element in it can have two kinds of format.

    - dict

    servers = [
        {'name':'node1','host':'127.0.0.1','port':10000,'db':0},
        {'name':'node2','host':'127.0.0.1','port':11000,'db':0},
        {'name':'node3','host':'127.0.0.1','port':12000,'db':0},
        ]


    - url_schema

    servers = ['redis://127.0.0.1:10000/0?name=node1',
                'redis://127.0.0.1:11000/0?name=node2',
                'redis://127.0.0.1:12000/0?name=node3'
        ]

    """
    configs = []
    if not isinstance(servers, list):
        raise ValueError("server's config must be list")
    _type = type(servers[0])
    if _type == dict:
        return servers
    if (sys.version_info[0] == 3 and _type in [str, bytes]) \
            or (sys.version_info[0] == 2 and _type in [str, unicode]):
        for config in servers:
            configs.append(parse_url(config))
    else:
        raise ValueError("invalid server config")
    return configs
